<?php
/**
 * PHPTAL templating engine
 *
 * PHP Version 5
 *
 * @category HTML
 * @package  PHPTAL
 * @author   Laurent Bedubourg <lbedubourg@motion-twin.com>
 * @author   Kornel Lesiński <kornel@aardvarkmedia.co.uk>
 * @license  http://www.gnu.org/licenses/lgpl.html GNU Lesser General Public License
 * @version  SVN: $Id: DocumentBuilder.php 833 2010-01-24 20:46:40Z kornel $
 * @link     http://phptal.org/
 */

namespace pear2\HTML\Template\PHPTAL\Dom; // PHPTAL_Dom_
use pear2\HTML\Template\PHPTAL as P;


require_once 'PHPTAL/Dom/SaxXmlParser.php';
require_once 'PHPTAL/Dom/XmlnsState.php';

require_once 'PHPTAL/Dom/Node.php';
require_once 'PHPTAL/Dom/XmlDeclaration.php';
require_once 'PHPTAL/Dom/DocumentType.php';
require_once 'PHPTAL/Dom/Text.php';
require_once 'PHPTAL/Dom/Element.php';
require_once 'PHPTAL/Dom/Comment.php';
require_once 'PHPTAL/Dom/CDATASection.php';
require_once 'PHPTAL/Dom/ProcessingInstruction.php';

/**
 * DOM Builder
 *
 * @package PHPTAL
 * @subpackage Dom
 */
class DocumentBuilder
{
    public function __construct()
    {
        $this->_xmlns = new P\Dom\XmlnsState(array(), '');
    }

    public function getResult()
    {
        return $this->documentElement;
    }

    public function getXmlnsState()
    {
        return $this->_xmlns;
    }

    // ~~~~~ XmlParser implementation ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    public function onDocumentStart()
    {
        $this->documentElement = new P\Dom\Element('documentElement', 'http://xml.zope.org/namespaces/tal',array(), $this->getXmlnsState());
        $this->documentElement->setSource($this->file, $this->line);
        $this->_stack = array();
        $this->_current = $this->documentElement;
    }

    public function onDocumentEnd()
    {
        if (count($this->_stack) > 0) {
            $left='</'.$this->_current->getQualifiedName().'>';
            for($i = count($this->_stack)-1; $i>0; $i--) $left .= '</'.$this->_stack[$i]->getQualifiedName().'>';
            throw new P\ParserException("Not all elements were closed before end of the document. Missing: ".$left);
        }
    }

    public function onDocType($doctype)
    {
        $this->pushNode(new P\Dom\DocumentType($doctype, $this->encoding));
    }

    public function onXmlDecl($decl)
    {
        $this->pushNode(new P\Dom\XmlDeclaration($decl, $this->encoding));
    }

    public function onComment($data)
    {
        $this->pushNode(new P\Dom\Comment($data, $this->encoding));
    }

    public function onCDATASection($data)
    {
        $this->pushNode(new P\Dom\CDATASection($data, $this->encoding));
    }

    public function onProcessingInstruction($data)
    {
        $this->pushNode(new P\Dom\ProcessingInstruction($data, $this->encoding));
    }

    public function onElementStart($element_qname, array $attributes)
    {
        $this->_xmlns = $this->_xmlns->newElement($attributes);

        if (preg_match('/^([^:]+):/', $element_qname, $m)) {
            $namespace_uri = $this->_xmlns->prefixToNamespaceURI($m[1]);
            if (false === $namespace_uri) {
                throw new P\ParserException("There is no namespace declared for prefix of element < $element_qname >");
            }
        } else {
            $namespace_uri = $this->_xmlns->getCurrentDefaultNamespaceURI();
        }

        $attrnodes = array();
        foreach ($attributes as $qname=>$value) {
            $local_name = $qname;
            if (preg_match('/^([^:]+):(.+)$/', $qname, $m)) {
                $local_name = $m[2];
                $attr_namespace_uri = $this->_xmlns->prefixToNamespaceURI($m[1]);
                if (false === $attr_namespace_uri) {
                    throw new P\ParserException("There is no namespace declared for prefix of attribute $qname of element < $element_qname >");
                }
            } else {
                $attr_namespace_uri = ''; // default NS. Attributes don't inherit namespace per XMLNS spec
            }

            if ($this->_xmlns->isHandledNamespace($attr_namespace_uri) 
                && !$this->_xmlns->isValidAttributeNS($attr_namespace_uri, $local_name)) {
                throw new P\ParserException("Attribute '$local_name' is in '$attr_namespace_uri' namespace, but is not a supported PHPTAL attribute");
            }

            $attrnodes[] = new P\Dom\Attr($qname, $attr_namespace_uri, $value, $this->encoding);
        }

        $node = new P\Dom\Element($element_qname, $namespace_uri, $attrnodes, $this->getXmlnsState());
        $this->pushNode($node);
        $this->_stack[] =  $this->_current;
        $this->_current = $node;
    }

    public function onElementData($data)
    {
        $this->pushNode(new P\Dom\Text($data, $this->encoding));
    }

    public function onElementClose($qname)
    {
        if ($this->_current === $this->documentElement) {
            throw new P\ParserException("Found closing tag for < $qname > where there are no open tags");
        }
        if ($this->_current->getQualifiedName() != $qname) {
            throw new P\ParserException("Tag closure mismatch, expected < /".$this->_current->getQualifiedName()." > (opened in line ".$this->_current->getSourceLine().") but found < /".$qname." >");
        }
        $this->_current = array_pop($this->_stack);
        if ($this->_current instanceof P\Dom\Element) {
            $this->_xmlns = $this->_current->getXmlnsState(); // restore namespace prefixes info to previous state
        }
    }

    private function pushNode(P\Dom\Node $node)
    {
        $node->setSource($this->file, $this->line);
        $this->_current->appendChild($node);
    }

    public function setSource($file, $line)
    {
        $this->file = $file; $this->line = $line;
    }

    public function setEncoding($encoding)
    {
        $this->encoding = $encoding;
    }

    private $file, $line;

    private $encoding;
    private $documentElement;    /* PHPTAL_Dom_Element */
    private $_stack;   /* array<PHPTAL_Dom_Node> */
    private $_current; /* PHPTAL_Dom_Node */
    private $_xmlns;   /* PHPTAL_Dom_XmlnsState */
}

