<?php
/**
 * PHPTAL templating engine
 *
 * PHP Version 5
 *
 * @category HTML
 * @package  PHPTAL
 * @author   Laurent Bedubourg <lbedubourg@motion-twin.com>
 * @author   Kornel Lesi?ski <kornel@aardvarkmedia.co.uk>
 * @license  http://www.gnu.org/licenses/lgpl.html GNU Lesser General Public License
 * @version  SVN: $Id: State.php 811 2009-12-10 17:54:41Z kornel $
 * @link     http://phptal.org/
 */

require_once 'PHPTAL/Php/Gen/XML.php';
require_once 'PHPTAL/Php/Gen/XHTML.php';
require_once 'PHPTAL/Php/Gen/HTML5.php';

/**
 * Base class for element hierarchy visitor that compiles DOM to PHP code.
 *
 * Its subclasses handle differences between rendering modes: XML, XHTML and HTML5.
 */
abstract class PHPTAL_Php_Gen_Standard
{
    protected $codewriter;

    public function __construct(PHPTAL_Php_CodeWriter $codewriter)
    {
        $this->codewriter = $codewriter;
    }

    /**
     * Generate PHP code that writes XML declaration.
     *
     * @param PHPTAL_Dom_XmlDeclaration $node
     * @return void
     */
    public function visitXmlDeclaration(PHPTAL_Dom_XmlDeclaration $node)
    {
        $this->codewriter->setXmlDeclaration($node->getValueEscaped());
        $this->codewriter->doXmlDeclaration();
    }

    /**
     * Generate PHP code that writes document type definition.
     *
     * @param PHPTAL_Dom_DocumentType $node
     * @return void
     */
    public function visitDocumentType(PHPTAL_Dom_DocumentType $node)
    {
        $this->codewriter->setDocType($node->getValueEscaped());
        $this->codewriter->doDoctype();
    }

    /**
     * Generate PHP code that writes processing instruction.
     *
     * @param PHPTAL_Dom_ProcessingInstruction $node
     * @return void
     */
    public function visitProcessingInstruction(PHPTAL_Dom_ProcessingInstruction $node)
    {
        if (preg_match('/^<\?(?:php|[=\s])/i', $node->getValueEscaped())) {
            // block will be executed as PHP
            $this->codewriter->pushHTML($node->getValueEscaped());
        } else {
            $this->codewriter->doEchoRaw("'<'");
            $this->codewriter->pushHTML(substr($this->codewriter->interpolateHTML($node->getValueEscaped()), 1));
        }
    }

    /**
     * Generate PHP code that writes comment.
     *
     * @param PHPTAL_Dom_Comment $node
     * @return void
     */
    public function visitComment(PHPTAL_Dom_Comment $node)
    {
        if (!preg_match('/^\s*!/', $node->getValueEscaped())) {
            $this->codewriter->pushHTML('<!--'.$node->getValueEscaped().'-->');
        }
    }

    /**
     * Generate PHP code that writes CDATA section.
     *
     * @param PHPTAL_Dom_CDATASection $node
     * @return void
     */
    public function visitCDATA(PHPTAL_Dom_CDATASection $node)
    {
        if ($node->getValueEscaped() !== '') {
             $this->codewriter->pushHTML($this->codewriter->interpolateCDATA('<![CDATA['.$node->getValueEscaped().']]>'));
        }
    }

    /**
     * Generate PHP code that writes text.
     *
     * @param PHPTAL_Dom_Text $node
     * @return void
     */
    public function visitText(PHPTAL_Dom_Text $node)
    {
        if ($node->getValueEscaped() !== '') {
            $this->codewriter->pushHTML($this->codewriter->interpolateHTML($node->getValueEscaped()));
        }
    }

    /**
     * Generate PHP code that writes the specified element, recursively.
     *
     * @param PHPTAL_Dom_Element $node
     * @return void
     */
    public function visitElement(PHPTAL_Dom_Element $node)
    {
        try {
            /// self-modifications

            $this->initElement($node);

            /// code generation

            if ($this->codewriter->isDebugOn()) {
                $this->codewriter->doSetVar('$ctx->_line', $node->getSourceLine());
                $this->codewriter->doComment('tag "'.$node->getQualifiedName().'" from line '.$node->getSourceLine());
            }

            $this->generateSurroundHead($node);

            if (count($node->replaceAttributes)) {
                foreach ($node->replaceAttributes as $att) {
                    $att->before($this->codewriter);
                    $att->after($this->codewriter);
                }
            } elseif (!$node->hidden) {
                // a surround tag may decide to hide us (tal:define for example)
                $this->generateHead($node);
                $this->generateContent($node);
                $this->generateFoot($node);
            }

            $this->generateSurroundFoot($node);
        }
        catch (PHPTAL_TemplateException $ex) {
            $ex->hintSrcPosition($node->getSourceFile(), $node->getSourceLine());
            throw $ex;
        }
    }

    /**
     * Preprocess element before generating its PHP code.
     *
     * @param PHPTAL_Dom_Element $node
     * @return void
     */
    protected function initElement(PHPTAL_Dom_Element $node)
    {
        $this->replacePHPAttributes($node);
    }

    /**
     * support <?php ?> inside attributes
     */
    protected function replacePHPAttributes(PHPTAL_Dom_Element $node)
    {
        foreach ($node->getAttributeNodes() as $attr) {
            $split = preg_split("/<\?(php|=|)(.*?)\?>/", $attr->getValueEscaped(), null, PREG_SPLIT_DELIM_CAPTURE);
            if (count($split) == 1) continue;

            $new_value = '';
            for ($i = 0; $i < count($split); $i += 3) {
                if (strlen($split[$i])) {
                    $new_value .= 'echo \''.str_replace('\'', '\\\'', $split[$i]).'\';';
                }

                if (isset($split[$i + 2])) {
                    if ($split[$i + 1] === '=') $new_value .= 'echo ';
                    $new_value .= rtrim($split[$i + 2], "; \n\r").';';
                }
            }
            $attr->overwriteValueWithCode($new_value);
        }
    }

    // ~~~~~ Generation methods may be called by some PHPTAL attributes ~~~~~

    public function generateSurroundHead(PHPTAL_Dom_Element $node)
    {
        foreach ($node->surroundAttributes as $att) {
            $att->before($this->codewriter);
        }
    }

    public function generateHead(PHPTAL_Dom_Element $node)
    {
        if ($node->headFootDisabled)
            return;

        if ($node->headPrintCondition) {
            $this->codewriter->doIf($node->headPrintCondition);
        }

        $this->openTag($node);

        if ($node->headPrintCondition) {
            $this->codewriter->doEnd('if');
        }
    }

    public function generateContent(PHPTAL_Dom_Element $node, $realContent = false)
    {
        if (!$this->isEmptyNode($node)) {
            if ($realContent || !count($node->contentAttributes)) {
                // Call recursively for each child node.
                foreach ($node->childNodes as $child) {
                    $child->accept($this);
                }
            }
            else foreach ($node->contentAttributes as $att) {
                $att->before($this->codewriter);
                $att->after($this->codewriter);
            }
        }
    }

    public function generateFoot(PHPTAL_Dom_Element $node)
    {
        if ($node->headFootDisabled)
            return;
        if ($this->isEmptyNode($node))
            return;

        if ($node->footPrintCondition) {
            $this->codewriter->doIf($node->footPrintCondition);
        }

        $this->closeTag($node);

        if ($node->footPrintCondition) {
            $this->codewriter->doEnd('if');
        }
    }

    public function generateSurroundFoot(PHPTAL_Dom_Element $node)
    {
        for ($i = (count($node->surroundAttributes)); $i > 0; $i--) {
            $node->surroundAttributes[$i - 1]->after($this->codewriter);
        }
    }

    protected function openTag(PHPTAL_Dom_Element $node)
    {
        $this->codewriter->pushHTML('<'.$node->getQualifiedName());

        $this->generateAttributes($node);

        if ($this->isEmptyNode($node)) {
            $this->codewriter->pushHTML('/>');
        } else {
            $this->codewriter->pushHTML('>');
        }
    }

    protected function closeTag(PHPTAL_Dom_Element $node)
    {
        $this->codewriter->pushHTML('</'.$node->getQualifiedName().'>');
    }

    protected function generateAttributes(PHPTAL_Dom_Element $node)
    {
        foreach ($node->getAttributeNodes() as $attr) {
            $this->generateAttribute($attr);
        }
    }

    protected function generateAttribute(PHPTAL_Dom_Attr $attr)
    {
        switch ($attr->getReplacedState()) {
            case PHPTAL_Dom_Attr::NOT_REPLACED:
                $this->codewriter->pushHTML(' '.$attr->getQualifiedName());
                $html = $this->codewriter->interpolateHTML($attr->getValueEscaped());
                $this->codewriter->pushHTML('='.$this->quoteAttributeValue($html));
                break;

            case PHPTAL_Dom_Attr::HIDDEN:
                break;

            case PHPTAL_Dom_Attr::FULLY_REPLACED:
                $this->codewriter->pushHTML($attr->getValueEscaped());
                break;

            case PHPTAL_Dom_Attr::VALUE_REPLACED:
                $this->codewriter->pushHTML(' '.$attr->getQualifiedName().'="');
                $this->codewriter->pushHTML($attr->getValueEscaped());
                $this->codewriter->pushHTML('"');
                break;
        }
    }

    public function quoteAttributeValue($value)
    {
        return '"'.$value.'"';
    }

    public function isEmptyNode(PHPTAL_Dom_Element $node)
    {
        return count($node->childNodes) == 0 && count($node->contentAttributes) == 0;
    }
}
